      program main 
!***********************************************************
! program to solve a finite difference 
! discretization of Helmholtz equation :  
! (d2/dx2)u + (d2/dy2)u - alpha u = f 
! using Jacobi iterative method. 
!
! Modified: Dieter an Mey,     Aachen University (RWTH), 1999 - 2003
! Modified: Sanjiv Shah,       Kuck and Associates, Inc. (KAI), 1998
! Author:   Joseph Robicheaux, Kuck and Associates, Inc. (KAI), 1998
! 
! Directives are used in this code to achieve paralleism. 
! All do loops are parallized with default 'static' scheduling.
! 
! Input :  n - grid dimension in x direction 
!          m - grid dimension in y direction
!          alpha - Helmholtz constant (always greater than 0.0)
!          tol   - error tolerance for iterative solver
!          relax - Successice over relaxation parameter
!          mits  - Maximum iterations for iterative solver
!
! On output 
!       : u(n,m) - Dependent variable (solutions)
!       : f(n,m) - Right hand side function 
!************************************************************
      implicit none 

      integer n,m,mits, kreps,nreps
      double precision tol,relax,alpha 
      integer nthreads, minthreads, maxthreads, omp_get_max_threads, omp_get_num_threads

      common /idat/ n,m,mits
      common /fdat/ tol,alpha,relax
! 
! Read info 
! 
      write(*,*) "Input minthreads,maxthreads - number of OpenMP threads " 
      read(*,*) minthreads,maxthreads
      write(*,*) "Input nreps - number of repetitions " 
      read(*,*) nreps
      write(*,*) "Input n,m - grid dimension in x,y direction " 
      read(*,*) n,m 
      write(*,*) "Input alpha - Helmholtz constant " 
      read(*,*) alpha
      write(*,*) "Input relax - Successive over-relaxation parameter"
      read(*,*) relax 
      write(*,*) "Input tol - error tolerance for iterative solver" 
      read(*,*) tol 
      write(*,*) "Input mits - Maximum iterations for solver" 
      read(*,*) mits
      
      write(*,*) "Input echo:"
      write(*,*) "minthreads,maxthreads - number of OpenMP threads ", minthreads,maxthreads
      write(*,*) "nreps - number of repetitions ", nreps
      write(*,*) "n,m - grid dimension in x,y direction ", n,m
      write(*,*) "alpha - Helmholtz constant ", alpha
      write(*,*) "relax - Successive over-relaxation parameter", relax
      write(*,*) "tol - error tolerance for iterative solver", tol 
      write(*,*) "mits - Maximum iterations for solver", mits

!
! Calls a driver routine 
! 
      minthreads = 1
!$    maxthreads = min(maxthreads, omp_get_max_threads() )
      
      do nthreads = minthreads, maxthreads
      do kreps = 1, nreps

      write(*,*) "nthreads - Number of OpenMP Threads", nthreads

!$    call omp_set_num_threads(nthreads)
!$    write(*,*) "nthreads - Number of OpenMP Threads", omp_get_num_threads()
      call driver

      end do
      end do

      stop
      end 

      subroutine driver
!************************************************************
! Subroutine driver () 
! This is where the arrays are allocated and initialzed. 
!
! Working varaibles/arrays 
!     dx  - grid spacing in x direction 
!     dy  - grid spacing in y direction 
!************************************************************
      implicit none 

      integer n,m,mits 
      real*8 realtime, t0, t1, t2
      double precision tol,relax,alpha, dt

      common /idat/ n,m,mits
      common /fdat/tol,alpha,relax

      double precision u(n,m),f(n,m),dx,dy
      double precision mflops, maxmflops

! Initialize data

      call initialize (n,m,alpha,dx,dy,u,f)

! Solve Helmholtz equation

      t0 = realtime()
      t1 = realtime()
      call jacobi (n,m,dx,dy,alpha,relax,u,f,tol,mits)
      t2 = realtime()
      dt = ( t2 - t1 )
      mflops = float(mits)*float(m-2)*float(n-2)*13.0d0*0.000001d0 / dt
      write (*,'(a,f12.6)') 'elapsed time usage', dt
      write (*,'(a,f12.6)') 'MFlop/s            ', mflops
     
! Check error between exact solution

      call error_check (n,m,alpha,dx,dy,u,f)

      return 
      end 

      subroutine initialize (n,m,alpha,dx,dy,u,f) 
!*****************************************************
! Initializes data 
! Assumes exact solution is u(x,y) = (1-x^2)*(1-y^2)
!
!*****************************************************
      implicit none 
     
      integer n,m
      double precision u(n,m),f(n,m),dx,dy,alpha
      
      integer i,j, xx,yy
      double precision PI 
      parameter (PI=3.1415926)

      dx = 2.0 / (n-1)
      dy = 2.0 / (m-1)

! Initilize initial condition and RHS

!$omp parallel do private(xx,yy,i,j)
      do j = 1,m
         do i = 1,n
            xx = -1.0 + dx * dble(i-1)        ! -1 < x < 1
            yy = -1.0 + dy * dble(j-1)        ! -1 < y < 1
            u(i,j) = 0.0 
            f(i,j) = -alpha *(1.0-xx*xx)*(1.0-yy*yy) &
     &           - 2.0*(1.0-xx*xx)-2.0*(1.0-yy*yy)
         enddo
      enddo


      return 
      end 

      subroutine error_check (n,m,alpha,dx,dy,u,f) 
      implicit none 
!***********************************************************
! Checks error between numerical and exact solution 
!
!*********************************************************** 
     
      integer n,m
      double precision u(n,m),f(n,m),dx,dy,alpha 
      
      integer i,j
      double precision xx,yy,temp,error 

      dx = 2.0 / (n-1)
      dy = 2.0 / (m-1)
      error = 0.0 

!$omp parallel do private(xx,yy,i,j,temp) reduction(+:error)
      do j = 1,m
         do i = 1,n
            xx = -1.0d0 + dx * dble(i-1)
            yy = -1.0d0 + dy * dble(j-1)
            temp  = u(i,j) - (1.0-xx*xx)*(1.0-yy*yy)
            error = error + temp*temp 
         enddo
      enddo

      error = sqrt(error)/dble(n*m)

      print *, 'Solution Error : ',error

      return 
      end 
